
// created 06.2003 by Stefan Kleine Stegemann
// 
// licensed under GPL

#include "ImageView.h"
#include "TransformableImageView.h"

#include <Foundation/NSString.h>
#include <Foundation/NSException.h>
#include <AppKit/NSClipView.h>
#include <AppKit/NSImage.h>
#include <AppKit/NSTextField.h>
#include <AppKit/NSScroller.h>
#include <AppKit/NSFont.h>
#include <AppKit/NSImageView.h>


/*
 * Non-Public methods.
 */
@interface ImageView(Private)
- (void) createScrollView;
- (void) createInternalImageView;
- (void) createMessageView;
- (void) centerImageInScrollView;
- (void) updateScrollers;
@end

/*
 * This view displays an image within a ScrollView. Additionaly,
 * some controls are provided to manipulate the display properties
 * (eg. the zoom factor).
 */
@implementation ImageView

/*
 * Designated initializer.
 */
- (id) initWithFrame: (NSRect)frame
{
   if (self = [super initWithFrame: frame])
   {
      [self createScrollView];
      [self createInternalImageView];
      [self createMessageView];

      [self setAutoresizesSubviews: YES];
      [self setAutoresizingMask: (NSViewWidthSizable |
                                  NSViewHeightSizable)];

      messageDisplayed = NO;
   }

   return self;
}


- (void) dealloc
{
   NSLog(@"dealloc ImageView, retain count is %d", [self retainCount]);

   [scrollView release];
   [internalImageView release];
   [messageView release];
   
   [super dealloc];
}


/*
 * When resized, center the image in the scrollview.
 */
- (void) resizeWithOldSuperviewSize: (NSSize)oldBoundsSize 
{
   [super resizeWithOldSuperviewSize: oldBoundsSize];
   //[self centerImageInScrollView];
   //[self updateScrollers];
}


/*
 * Center the image in it's scrollview when necessary.
 */
- (void) centerImageInScrollView
{
   NSRect scrollViewFrame;
   NSSize imageSize;
   NSRect imageFrame;
   BOOL   needsResize = NO;

   //NSLog(@"center image");
   scrollViewFrame = [scrollView documentVisibleRect];
   imageFrame      = [internalImageView frame];

   imageSize = [internalImageView frame].size;

   if (scrollViewFrame.size.width > imageSize.width)
   {
      imageFrame.size.width = scrollViewFrame.size.width;
      needsResize = YES;
   }
   else if (scrollViewFrame.size.width < imageSize.width)
   {
      imageFrame.size.width = imageSize.width;
      needsResize = YES;
   }

   if (scrollViewFrame.size.height > imageSize.height)
   {
      imageFrame.size.height = scrollViewFrame.size.height;
      needsResize = YES;
   }
   else if (scrollViewFrame.size.height < imageSize.height)
   {
      imageFrame.size.height = imageSize.height;
      needsResize = YES;
   }
      
   
   if (needsResize)
   {
      /*
      NSLog(@"   x:%f y:%f w:%f h:%f",
            imageFrame.origin.x,
            imageFrame.origin.y,
            imageFrame.size.width,
            imageFrame.size.height);
      */
   
      [internalImageView setFrame: imageFrame];
   }
}


/*
 * Determine whether Scrollers should be visible or
 * not. This depends on the images size.
 */
- (void) updateScrollers
{
   NSSize scrollViewSize;
   NSSize imageViewSize;

   //scrollViewSize = [scrollView documentVisibleRect].size;
   scrollViewSize = [scrollView contentSize];
   imageViewSize  = [internalImageView frame].size;

   if (![scrollView displayPageControls])
   {
      [scrollView setHasHorizontalScroller: scrollViewSize.width < imageViewSize.width];
   }
   else
   {
      [scrollView setHasHorizontalScroller: YES];
   }

   [scrollView setHasVerticalScroller: scrollViewSize.height < imageViewSize.height];
}


/*
 * Set the image to be displayed.
 */
- (void) setImage: (NSImage*)anImage
{
   [self removeMessage];

   [internalImageView setImage: anImage];
}


/*
 * Repaint the the view, update the scrollers ...
 */
- (void) updateView
{
   //[self updateScrollers];
   [internalImageView setNeedsDisplay: YES];
   [self setNeedsDisplay: YES];
}


/*
 * Get the embedded image scroll view.
 */
- (ImageScrollView*) scrollView
{
   return scrollView;
}


/*
 * Display a message in the internal ImageView.
 * This can be used to display informations for
 * the user (eg. when work is progress). The
 * message will disapper when removeMessage
 * or setImage is called.
 */
- (void) displayMessage: (NSString*)message
{
   NSRect mFrame;
   NSRect iFrame;

   [messageView setStringValue: message];
   [messageView sizeToFit];
   
   iFrame = [internalImageView frame];
   mFrame = [messageView frame];

   mFrame = NSMakeRect(NSMidX(iFrame) - (NSWidth(mFrame) / 2),
                       NSMidY(iFrame) - (NSHeight(mFrame) / 2),
                       NSWidth(mFrame),
                       NSHeight(mFrame));

   [messageView setFrame: mFrame];
                             
   [internalImageView addSubview: messageView];
   messageDisplayed = YES;
}


/*
 * Remove a displayed message from the ImageView.
 * Use this only if displayMessage has been
 * called before.
 */
- (void) removeMessage
{
   if (messageDisplayed)
   {
      [messageView removeFromSuperview];
      messageDisplayed = NO;
   }
}


/*
 * Resize the internal ImageView so that is shows the
 * complete image.
 */
- (void) sizeToFitImage
{
   NSSize imageSize;
   NSRect viewFrame;

   imageSize = [[internalImageView image] size];
   viewFrame = NSMakeRect(0, 0, imageSize.width, imageSize.height);

   [internalImageView setFrame: viewFrame];
}


/*
 * Set the size of the internal image view. This can be used
 * to scale the displayed image becaue  the internal image view
 * is configured to scale it's image when resized.
 */
- (void) setImageSize: (NSSize)size
{
   NSRect viewFrame;

   viewFrame = NSMakeRect(0, 0, size.width, size.height);
   [internalImageView setFrame: viewFrame];
}


/*
 * Create the ScrollView that holds the image.
 */
- (void) createScrollView
{
   NSRect scrollViewFrame;

   scrollViewFrame = [self frame];
   scrollView = [[ImageScrollView alloc] initWithFrame: scrollViewFrame];

   [scrollView setHasHorizontalScroller: YES];
   [scrollView setHasVerticalScroller: YES];
   [scrollView setAutoresizesSubviews: YES];
   [scrollView setAutoresizingMask: (NSViewWidthSizable |
                                     NSViewHeightSizable)];
   [scrollView setDrawsBackground: NO];
   
   [self addSubview: scrollView];
}


- (void) createInternalImageView
{
   NSRect   frame;

   frame = NSMakeRect(0, 0,
                      [scrollView contentSize].width,
                      [scrollView contentSize].height);
   //internalImageView = [[TransformableImageView alloc] initWithFrame: frame];
   internalImageView = [[NSImageView alloc] initWithFrame: frame];
   [internalImageView setImageAlignment: NSImageAlignCenter];
   [internalImageView setImageScaling: NSScaleNone];
   [internalImageView setImageScaling: NSScaleProportionally];

   [scrollView setDocumentView: internalImageView];
}


- (void) createMessageView
{
   NSRect   frame;

   frame = NSMakeRect(0, 0, 0, 0);
   messageView = [[NSTextField alloc] initWithFrame: frame];
   [messageView setFont: [NSFont boldSystemFontOfSize: 16.0]];
   [messageView setBordered: NO];
   [messageView setBezeled: NO];
   [messageView setSelectable: NO];
   [messageView setDrawsBackground: NO];
   [messageView setEditable: NO];
}


@end
