
// created 06.2003 by Stefan Kleine Stegemann
// 
// licensed under GPL

#include "ImageWindow.h"
#include "ImageWindowController.h"

#include <Foundation/NSString.h>
#include <Foundation/NSException.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSEnumerator.h>
#include <AppKit/NSButton.h>
#include <AppKit/NSImage.h>
#include <AppKit/NSTextField.h>
#include <AppKit/NSFont.h>
#include <AppKit/NSProgressIndicator.h>


/*
 * Non-Public methods of the ImageView class.
 */
@interface ImageWindow(Private)
- (void) createContentView;
- (void) createImageView;
- (void) createToolbar;
- (void) rearrangeToolbarButtons: (BOOL)displayPageNav;
- (NSButton*) makeToolbarButtonWithTitle: (NSString*)title
                                   image: (NSString*)imageName;
- (void) makeToolbarProgressBar;
- (void) addToToolbar: (id)aView
           atPosition: (int)pos;
@end


/*
 * A window that displays a ViewableDocument and provides
 * some controls to control the view (eg. zoom controls).
 * An ImageWindow is controlled by an instance of
 * DocumentController. 
 */
@implementation ImageWindow

/*
 * Designated initializer.
 */
- (id) init
{
   NSRect  contentRect;

   contentRect = NSMakeRect(100, 100, 400, 200);

   if (self = [super initWithContentRect: contentRect
                     styleMask: (NSTitledWindowMask
                                 | NSClosableWindowMask
                                 | NSMiniaturizableWindowMask
                                 | NSResizableWindowMask)
                     backing: NSBackingStoreBuffered
                     defer: NO])
   {
      myController = nil;

      [self setTitle: @"ViewIt Window"];
      [self createContentView];
      [self createImageView];
      [self createToolbar];

      [self setDisplayPageNavigation: NO];
   }

   return self;
}


/*
 * Set the windows controller.
 */
- (void) setController: (ImageWindowController*)controller
{
   myController = controller;
   // myController needs not to be retained because this
   // is done by the document-based application model

   [[[imageView scrollView] firstPageButton] setAction: @selector(firstPage:)];
   [[[imageView scrollView] firstPageButton] setTarget: myController];
   [[[imageView scrollView] prevPageButton] setAction: @selector(previousPage:)];
   [[[imageView scrollView] prevPageButton] setTarget: myController];
   [[[imageView scrollView] nextPageButton] setAction: @selector(nextPage:)];
   [[[imageView scrollView] nextPageButton] setTarget: myController];   
   [[[imageView scrollView] lastPageButton] setAction: @selector(lastPage:)];
   [[[imageView scrollView] lastPageButton] setTarget: myController];
   [[[imageView scrollView] pageField] setDelegate: myController];

   [tbPrevPage setAction: @selector(previousPage:)];
   [tbPrevPage setTarget: myController];
   [tbNextPage setAction: @selector(nextPage:)];
   [tbNextPage setTarget: myController];
   [tbZoomIn setAction: @selector(zoomIn:)];
   [tbZoomIn setTarget: myController];
   [tbZoomOut setAction: @selector(zoomOut:)];
   [tbZoomOut setTarget: myController];
   [tbZoomToRealSize setAction: @selector(zoomToRealSize:)];
   [tbZoomToRealSize setTarget: myController];   
   [tbSizeToFit setAction: @selector(sizePageToFit:)];
   [tbSizeToFit setTarget: myController];
   [tbSizeToFitWidth setAction: @selector(sizePageToFitWidth:)];
   [tbSizeToFitWidth setTarget: myController];   
}


- (void) dealloc
{
   NSLog(@"dealloc ImageWindow, retain count is %d", [self retainCount]);

   [[self contentView] release];
   [imageView release];
   [toolbar release];

   [tbNextPage release];
   [tbPrevPage release];
   [tbZoomIn release];
   [tbZoomOut release];
   [tbZoomToRealSize release];
   [tbSizeToFit release];
   [tbSizeToFitWidth release];
   [tbProgressBar release];

   [tbElements release];

   [super dealloc];
}


- (ImageView*) imageView
{
   return imageView;
}


- (NSProgressIndicator*) toolbarProgressBar
{
   return tbProgressBar;
}


/*
 * Sets whether controls to navigate through pages are
 * displayed or not. These controls are disabled by default.
 */
- (void) setDisplayPageNavigation: (BOOL)displayPageNav
{
   [[[self imageView] scrollView] setDisplayPageControls: displayPageNav];
   [self rearrangeToolbarButtons: displayPageNav];
}


/*
 * Create the windows content view.
 */
- (void) createContentView
{
   NSView* contentView;
   NSRect contentRect;
   NSRect contentViewFrame;

   contentRect = [[self contentView] frame];
   contentViewFrame = NSMakeRect(0, 0, 
                                 contentRect.size.width,
                                 contentRect.size.height);
   contentView = [[NSView alloc] initWithFrame: contentViewFrame];
   [contentView setAutoresizesSubviews: YES];
   [contentView setAutoresizingMask: (NSViewWidthSizable |
                                      NSViewHeightSizable)];
   
   [self setContentView: contentView];
}


/*
 * Create the image view and add it to the window.
 */
- (void) createImageView
{
   NSRect imageViewFrame;
   NSRect contentRect;

   contentRect = [[self contentView] frame];

   imageViewFrame = NSMakeRect(0, 0, 
                               contentRect.size.width,
                               contentRect.size.height - TOOLBAR_HEIGHT);
   imageView = [[ImageView alloc] initWithFrame: imageViewFrame];

   [[self contentView] addSubview: imageView];
}


/*
 * Create the window's toolbar. As long as NSToolbar is not
 * fully available on GNUstep, the toolbar will be a view
 * with some buttons.
 */
- (void) createToolbar
{
   NSRect contentRect;
   NSRect toolbarFrame;

   contentRect = [[self contentView] frame];

   toolbarFrame = NSMakeRect(0,
                             (contentRect.size.height - TOOLBAR_HEIGHT + 1),
                             contentRect.size.width,
                             TOOLBAR_HEIGHT);

   toolbar = [[NSView alloc] initWithFrame: toolbarFrame];
   [toolbar setAutoresizingMask: (NSViewMinYMargin | NSViewWidthSizable)];

   tbElements = [[NSMutableArray alloc] initWithCapacity: 0];

   tbPrevPage = [self makeToolbarButtonWithTitle: @"Previous"
                      image: @"tb_prev_page.tiff"];
   
   tbNextPage = [self makeToolbarButtonWithTitle: @"Next"
                      image: @"tb_next_page.tiff"];

   tbZoomIn = [self makeToolbarButtonWithTitle: @"Bigger"
                    image: @"zoom_in.tiff"];
   
   tbZoomOut = [self makeToolbarButtonWithTitle: @"Smaller"
                     image: @"zoom_out.tiff"];

   tbZoomToRealSize = [self makeToolbarButtonWithTitle: @"1:1"
                            image: @"zoom_11.tiff"];

   tbSizeToFit = [self makeToolbarButtonWithTitle: @"Fit Window"
                       image: @"fit_page.tiff"];

   tbSizeToFitWidth = [self makeToolbarButtonWithTitle: @"Fit Width"
                            image: @"fit_width.tiff"];

   [self makeToolbarProgressBar];

   [self rearrangeToolbarButtons: NO];

   [[self contentView] addSubview: toolbar];   
}


/*
 * Create the progress bar that can be displayed on the toolbar.
 */
- (void) makeToolbarProgressBar
{
   NSRect frame = NSMakeRect(1,
                             (TOOLBAR_HEIGHT / 2) - (TOOLBAR_HEIGHT / 8),
                             TOOLBAR_HEIGHT,
                             TOOLBAR_HEIGHT / 4);

   tbProgressBar = [[NSProgressIndicator alloc] initWithFrame: frame];

   [tbProgressBar setBezeled: YES];
   [tbProgressBar setIndeterminate: NO];
   [tbProgressBar setVertical: NO];
   [tbProgressBar setMinValue: 0];
}


/*
 * Create the buttons displayed on the toolbar. Buttons
 * for page navigation are shown depending on the value
 * of displayPageNav.
 */
- (void) rearrangeToolbarButtons: (BOOL)displayPageNav
{
   NSButton*     aButton;
   NSEnumerator* e;
   int           firstNonNavButton;

   e = [tbElements objectEnumerator];
   while (aButton = [e nextObject])
   {
      [aButton removeFromSuperview];
   }

   
   if (displayPageNav)
   {
      [self addToToolbar: tbPrevPage atPosition: 1];
      [self addToToolbar: tbNextPage atPosition: 2];
      firstNonNavButton = 3;
   }
   else
   {
      firstNonNavButton = 1;
   }
   
   [self addToToolbar: tbZoomIn atPosition: firstNonNavButton];
   [self addToToolbar: tbZoomOut atPosition: firstNonNavButton + 1];
   [self addToToolbar: tbZoomToRealSize atPosition: firstNonNavButton + 2];
   [self addToToolbar: tbSizeToFit atPosition: firstNonNavButton + 3];
   [self addToToolbar: tbSizeToFitWidth atPosition: firstNonNavButton + 4];

   //[self addToToolbar: tbProgressBar atPosition: firstNonNavButton + 5];

   [toolbar setNeedsDisplay: YES];
}


/*
 * Create a button that is shown on the toolbar. A toolbar
 * button will have an image and a text that is displayed
 * below the image.
 */
- (NSButton*) makeToolbarButtonWithTitle: (NSString*)title
                                   image: (NSString*)imageName
{
   NSRect frame;
   NSButton* theButton;

   frame = NSMakeRect(1,
                      1,
                      TOOLBAR_HEIGHT - 4,
                      TOOLBAR_HEIGHT - 4);

   theButton = [[NSButton alloc] initWithFrame: frame];
   [theButton setFont: [NSFont systemFontOfSize: 8.0]];
   [theButton setButtonType: NSMomentaryLight];
   [theButton setTitle: title];
   [theButton setImagePosition: NSImageAbove];
   [theButton setImage: [NSImage imageNamed: imageName]];
   [theButton setBordered: YES];
#ifdef MACOSX
   [theButton setBezelStyle: NSRegularSquareBezelStyle];
#endif

   return theButton;
}


/*
 * Add a button to the toolbar at the specified position.
 */
- (void) addToToolbar: (id)aView
           atPosition: (int)pos
{
   NSAssert(aView != nil, @"toolbar element is nil");

   NSRect frame = [aView frame];
   [aView setFrameOrigin: NSMakePoint(8.0 + (TOOLBAR_HEIGHT * (pos - 1)),
                                      frame.origin.y)];

   [toolbar addSubview: aView];
   [tbElements addObject: aView];
}


@end
