
// created 06.2003 by Stefan Kleine Stegemann
// 
// licensed under GPL


#include "AppController.h"
#include "ImageWindowController.h"
#include "ImageWindow.h"
#include "ViewableDocument.h"
#include "ViewablePage.h"
#include "PageCache.h"

#include <Foundation/NSException.h>
#include <Foundation/NSString.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSValue.h>
#include <Foundation/NSNotification.h>
#include <AppKit/NSApplication.h>
#include <AppKit/NSScreen.h>
#include <AppKit/NSScroller.h>
#include <AppKit/NSImage.h>
#include <AppKit/NSTextField.h>
#include <AppKit/NSClipView.h>
#include <AppKit/NSMenu.h>
#include <AppKit/NSMenuItem.h>
#include <AppKit/NSProgressIndicator.h>


/*
 * Non-Public methods of ImageWindowController
 */
@interface ImageWindowController(Private)
- (void) initWindow;
- (void) updatePageImage;
- (void) scrollToTopOfPage;
- (void) cacheProgressNotification: (NSNotification*)notification;
@end

/*
 * An instance of DocumentController controlls an
 * ImageWindow that displays a ViewableDocument. 
 */
@implementation ImageWindowController

- (id) initWithWindow: (NSWindow*)win
{
   if (self = [super initWithWindow: win])
   {
      [self setShouldCascadeWindows: NO];
      [(ImageWindow*)win setController: self];
      [self setScaleFactor: 1.0];
      documentIsDisplayed = NO;
      currentPaperSize = nil;
      currentPageOrientation = nil;
   }

   return self;
}


- (void) dealloc
{
   NSLog(@"dealloc ImageWindowController, retain count is %d", [self retainCount]);
   
   [[NSNotificationCenter defaultCenter] removeObserver: self];

   [super dealloc];
}


/*
 * Set the page to be displayed.
 */
- (void) setCurrentPage: (ViewablePage*)aPage
{
   id        win;
   NSString* pageFieldText;

   NSAssert(aPage != nil, @"new current Page is nil");
   NSAssert([self window] != nil, @"no window");

   currentPage = aPage;
   [self updatePageImage];

   pageFieldText = [[NSString alloc] initWithFormat: @"%d of %d",
                                     [currentPage pageNumber],
                                     [[self document] countPages]];
   [pageFieldText autorelease];

   win = [self window];
   [[[[win imageView] scrollView] pageField] setStringValue: pageFieldText];
}


/*
 * Get the page that is currently displayed.
 */
- (ViewablePage*)currentPage
{
   return currentPage;
}


/*
 * Set the scale factor (zoom).
 */
- (void) setScaleFactor: (double)factor
{
   scaleFactor = factor;
   NSLog(@"new scale factor is %f", scaleFactor);
}


/*
 * Get the current scale factor.
 */
- (double) scaleFactor
{
   return scaleFactor;
}


/*
 * Updates the managed ImageWindow to display the contents
 * of the controllers document. When invoked, a document
 * and a window must have been associated with this controller.
 */
- (void) initWindow
{
   id win;

   NSAssert([self window] != nil, @"no window");
   NSAssert([self document] != nil, @"no document");
   NSAssert([[self document] countPages] > 0, @"no pages");

   win = [self window];

   // set defaults for paper-size & page-orientation
   if ([[self document] canSetPapersize])
   {
      id psizeMenu = [[[[NSApplication sharedApplication] mainMenu] 
                         itemWithTitle: @"Papersize"] submenu];

      currentPaperSize = [psizeMenu itemWithTitle: DEFAULT_PAPERSIZE];
      currentPageOrientation = [psizeMenu itemWithTitle: PAGE_ORIENTATION_PORTRAIT];
      
      [currentPaperSize setState: NSOnState];
      [currentPageOrientation setState: NSOnState];
   }
   
   // display the document
   [self setCurrentPage: [[self document] pageAt: 1]];

   if ([[self document] countPages] > 1)
   {
      [win setDisplayPageNavigation: YES];
   }
   else
   {
      [win setDisplayPageNavigation: NO];
   }
   
   [self sizePageToMax];

   // get informed about page prefetching progress
   [[NSNotificationCenter defaultCenter]
      addObserver: self
      selector: @selector(cacheProgressNotification:)
      name: CACHE_PROGRESS_NOTIFICATION
      object: [self document]];   

   // tell the document that it is now displayed
   [[self document] documentIsDisplayed];
}


/*
 * Refreshes the currently displayed image from the
 * the current page. This is usefull for example when 
 * the image has been scaled and the view should 
 * display the new, scaled image.
 */
- (void) updatePageImage
{
   id win = [self window];

   NSAssert([self window] != nil, @"no window");
   NSAssert([self currentPage] != nil, @"no current page");

   if ([[self currentPage] canScale])
   {
      [[win imageView] setImage: [[self currentPage] imageScaledBy: [self scaleFactor]]];
      [[win imageView] sizeToFitImage];
   }
   else
   {
      NSSize scaledSize;
      
      scaledSize = 
         NSMakeSize([[[self currentPage] image] size].width * [self scaleFactor],
                    [[[self currentPage] image] size].height * [self scaleFactor]);

      [[win imageView] setImageSize: scaledSize];
      [[win imageView] setImage: [[self currentPage] image]];
   }

   [[win imageView] updateView];
}


/*
 * Resize window to show as much of the image as
 * possible. This is the natural size of the image 
 * but limited to the screen frame. After the window
 * has been resized, the image is scaled to fit in
 * the window.
 */
- (void) sizePageToMax
{
   NSSize  imageSize;
   NSSize  winSize;
   NSRect  screenFrame;
   id      win;
   float   maxAllowedHeight;
   float   maxAllowedWidth;

   NSAssert([self window] != nil, @"no window");
   NSAssert([self currentPage] != nil, @"current Page is nil");

   win = [self window];
   
   imageSize = [[[self currentPage] image] size];
   screenFrame = [[NSScreen mainScreen] visibleFrame];

   maxAllowedHeight = 
        screenFrame.size.height
      - 50
      - TOOLBAR_HEIGHT
      - [NSScroller scrollerWidthForControlSize: NSRegularControlSize];

   maxAllowedWidth =
        screenFrame.size.width
      - 50
      - [NSScroller scrollerWidthForControlSize: NSRegularControlSize];
      
   if (imageSize.height > maxAllowedHeight)
   {
      imageSize.height = maxAllowedHeight;
   }

   if (imageSize.width > maxAllowedWidth)
   {
      imageSize.width = maxAllowedWidth;
   }

   winSize = [self scalePageToFitIn: imageSize];

   winSize.height =
        winSize.height
      + [NSScroller scrollerWidthForControlSize: NSRegularControlSize]
      + TOOLBAR_HEIGHT
      + 2;
      
   winSize.width =
        winSize.width
      + [NSScroller scrollerWidthForControlSize: NSRegularControlSize]
      + 2;

   [self updatePageImage];
   [win setContentSize: winSize];  
}


/*
 * Go the first page of the document.
 */
- (void) firstPage: (id)sender
{
   NSAssert([self document] != nil, @"no document");

   [self setCurrentPage: [[self document] pageAt: 1]];
   [self scrollToTopOfPage];
}


/*
 * Go to the previous page, if available.
 */
- (void) previousPage: (id)sender
{
   int newPageNum;

   NSAssert([self document] != nil, @"no document");
   NSAssert([self currentPage] != nil, @"no current page");

   newPageNum = [[self currentPage] pageNumber] - 1;
   if (newPageNum > 0)
   {
      NSLog(@"current page is %d, new page is %d",
            [[self currentPage] pageNumber],
            newPageNum);

      [self setCurrentPage: [[self document] pageAt: newPageNum]];
      [self scrollToTopOfPage];
   }
}


/*
 * Go to the next page, if available.
 */
- (void) nextPage: (id)sender
{
   int newPageNum;

   NSAssert([self document] != nil, @"no document");
   NSAssert([self currentPage] != nil, @"no current page");

   newPageNum = [[self currentPage] pageNumber] + 1;
   if (newPageNum <= [[self document] countPages])
   {
      NSLog(@"current page is %d, new page is %d",
            [[self currentPage] pageNumber],
            newPageNum);

      [self setCurrentPage: [[self document] pageAt: newPageNum]];
      [self scrollToTopOfPage];
   }
}


/*
 * Go the first page of the document.
 */
- (void) lastPage: (id)sender
{
   NSAssert([self document] != nil, @"no document");

   [self setCurrentPage: [[self document] pageAt: [[self document] countPages]]];
   [self scrollToTopOfPage];
}


/*
 * Display the page in bigger size.
 */
- (void) zoomIn: (id)sender
{
   [self setScaleFactor: [self scaleFactor] + ZOOM_STEP];
   [self updatePageImage];
}


/*
 * Display the page in smaller size.
 */
- (void) zoomOut: (id)sender
{
   float newScaleFactor;

   newScaleFactor = [self scaleFactor] - ZOOM_STEP;

   if (newScaleFactor > 0.1)
   {
      [self setScaleFactor: newScaleFactor];
      [self updatePageImage];
   }
}


/*
 * Display the page in it's original size.
 */
- (void) zoomToRealSize: (id)sender
{
   [self setScaleFactor: 1.0];
   [self updatePageImage];
}


/*
 * Rescale the page so that it fits completly
 * in the window.
 */
- (void) sizePageToFit: (id)sender
{
   id     win;
   NSSize size;

   NSAssert([self window] != nil, @"no window");

   win = [self window];
   //   size = NSMakeSize([[[win imageView] scrollView] contentSize].width,
   //                  [[[win imageView] scrollView] contentSize].height);
   size = [[[win imageView] scrollView] contentSize];

   [self scalePageToFitIn: size];
   
   [self updatePageImage];
}


/*
 * Rescale the page so that it fits in the
 * window horizontally.
 */
- (void) sizePageToFitWidth: (id)sender
{
   id      win;
   NSSize  size;
   NSSize  newSize;
   double  oldScale;

   NSAssert([self window] != nil, @"no window");

   win = [self window];
   size = NSMakeSize([[[win imageView] scrollView] contentSize].width,
                     0);

   oldScale = [self scaleFactor];
   newSize = [self scalePageToFitIn: size];
   
   [self updatePageImage];
}


/*
 * If possible, set the documents papersize.
 */
- (void) setPaperSize: (id)sender
{
   NSAssert([self document] != nil, @"no document");

   if ([[self document] canSetPapersize])
   {
      NSDictionary* sizeDict;
      NSSize        size;

      NSLog(@"set papersize to %@", [sender title]);

      [currentPaperSize setState: NSOffState];
      [sender setState: NSOnState];
      currentPaperSize = sender;
      
      sizeDict = [[[AppController sharedController] papersizes] 
                    objectForKey: [sender title]];

      NSAssert(sizeDict != nil, @"Papersize not found");

      size = NSMakeSize([[sizeDict objectForKey: @"w"] floatValue],
                        [[sizeDict objectForKey: @"h"] floatValue]);

      NSLog(@"   papersize is %f x %f", size.width, size.height);

      [[self document] setPapersize: size];
      [self updatePageImage];
   }
   else
   {
      NSLog(@"cannot set the papersize for this document");
   }
}


/*
 * If possible, set the documents page orientation
 * (only if the documents papersize can be adjusted).
 */
- (void) setPageOrientation: (id)sender
{
   NSAssert([self document] != nil, @"no document");

   if ([[self document] canSetPapersize])
   {
      [currentPageOrientation setState: NSOffState];
      [sender setState: NSOnState];
      currentPageOrientation = sender;

      if ([[sender title] isEqualToString: PAGE_ORIENTATION_LANDSCAPE])
      {
         NSLog(@"set page orientation %@", PAGE_ORIENTATION_LANDSCAPE);
         [(ViewableDocument*)[self document] setPageOrientation: 
                                             PageOrientation_Landscape];
      }
      else if ([[sender title] isEqualToString: PAGE_ORIENTATION_PORTRAIT])
      {
         NSLog(@"set page orientation %@", PAGE_ORIENTATION_PORTRAIT);
         [(ViewableDocument*)[self document] setPageOrientation:
                                             PageOrientation_Portrait];
      }
      else
      {
         NSLog(@"invalid page orientation %@", [sender title]);
      }

      [self updatePageImage];
   }
}


/*
 * Get notified when the user changed the value
 * in the page field. When the user entered a valid
 * page number, this page is set as current page.
 */
- (void) controlTextDidEndEditing: (NSNotification*)aNotification
{
   int requestedPage;
   id  win;

   win = [self window];

   requestedPage = [[[[win imageView] scrollView] pageField] intValue];
   if ((requestedPage > 0) && (requestedPage <= [[self document] countPages]))
   {
      [self setCurrentPage: [[self document] pageAt: requestedPage]];
   }
}


/*
 * The Items of the "Papersizes" menu are only enabled if the
 * papersize of the document can be set.
 */
- (BOOL) validateMenuItem: (id)menuItem
{
   BOOL itemEnabled = YES;

   if ([[[menuItem menu] title] isEqualToString: @"Papersize"])
   {
      itemEnabled = [[self document] canSetPapersize];
   }

   return itemEnabled;
}


/*
 * Update the scale factor so that the image will fit into
 * a rectangle of the specified size. The new (scaled) size
 * is returned. Note that scaling is not performed until
 * the image is updated (updatePageImage).
 */
- (NSSize) scalePageToFitIn: (NSSize)size
{
   NSSize imageSize;
   NSSize newSize;
   double scaleFactorX;
   double scaleFactorY;
   double newScaleFactor;

   NSAssert([self currentPage] != nil, @"current Page is nil");

   imageSize = [[[self currentPage] image] size];
   // we need the real (unscaled) image size if the page
   // can scale
   if ([[self currentPage] canScale])
   {
      imageSize.width  = imageSize.width / [self scaleFactor];
      imageSize.height = imageSize.height / [self scaleFactor];
   }

   scaleFactorX = (size.width > 0 ? (size.width / imageSize.width) : 1000.0);
   scaleFactorY = (size.height > 0 ? (size.height / imageSize.height) : 1000.0);
   
   newScaleFactor = (scaleFactorX < scaleFactorY ? scaleFactorX : scaleFactorY);

   newSize = NSMakeSize(imageSize.width * newScaleFactor,
                        imageSize.height * newScaleFactor);

   [self setScaleFactor: newScaleFactor];

   return newSize;
}


/*
 * Scrolls to the top of the current page.
 */
- (void) scrollToTopOfPage
{
   id      win = [self window];
   NSPoint topOfPage;

   NSAssert([self currentPage] != nil, @"no current page");
   NSAssert([self window] != nil, @"no window");

   topOfPage =
      NSMakePoint(0,
                  [[[[win imageView] scrollView] documentView] frame].size.height);

   [[[[win imageView] scrollView] contentView] scrollToPoint: topOfPage];
}


/*
 * To be invoked when notifications about the cache loading
 * progress of the displayed document arrive.
 */
- (void) cacheProgressNotification: (NSNotification*)notification
{
   id             win = [self window];
   NSDictionary*  progressInfo;
   int            prefetchCount;
   int            pagesDone;

   NSAssert([win toolbarProgressBar] != nil, @"no progress indicator");

   progressInfo = [notification userInfo];
   pagesDone =
      [[progressInfo objectForKey: CACHE_PROGRESS_INFO_PAGES_DONE] intValue];
   prefetchCount =
      [[progressInfo objectForKey: CACHE_PROGRESS_INFO_PREFETCH_COUNT] intValue];

   NSLog(@"GOT CACHE PROGRESS NOTIFICATION (%d of %d)", pagesDone, prefetchCount);

   [[win toolbarProgressBar] setMaxValue: prefetchCount * 2];
   [[win toolbarProgressBar] setDoubleValue: pagesDone];
}


- (void) windowDidBecomeMain: (NSNotification*)aNotification
{
   if (([self document] != nil) && (!documentIsDisplayed))
   {
      documentIsDisplayed = YES;
      [self initWindow];
      [[self window] center];
   }
}


- (void) windowWillClose: (NSNotification*)aNotification
{
   NSLog(@"ImageWindow closed");
   NSLog(@"Retain count of ImageWindowController is %d", [self retainCount]);
   NSLog(@"Retain count of owned Document is %d", [[self document] retainCount]);

   // due to a GNUstep bug we have to autorlease the controller here
   // (see NSWindowController.m)
   [[self window] autorelease];
}

@end
