
// created 06.2003 by Stefan Kleine Stegemann
// 
// licensed under GPL

#include "ViewableDocument.h"
#include "ImageWindowController.h"
#include "AppController.h"
#include "ViewablePage.h"

#include <Foundation/NSException.h>
#include <Foundation/NSString.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSLock.h>
#include <Foundation/NSObject.h>


/*
 * Abstraction from something that can be viewed with 
 * ViewIt. ViewIt simply displays any ViewableDocument
 * that it gets. A ViewableDocument consists of a set
 * of pages, see ViewablePage for more informations.
 */
@implementation ViewableDocument

/*
 * designated initializer
 */
- (id) init
{
   if (self = [super init])
   {
      // a document has minimum 1 page
      pageStore = [[NSMutableArray arrayWithCapacity: 1] retain];
      pageStoreLock = [[NSLock alloc] init];
   }

   return self;
}


- (void) dealloc
{
   NSLog(@"dealloc ViewableDocument, retain count is %d", [self retainCount]);

   [pageStore release];
   [pageStoreLock release];
   
   [super dealloc];
}


// Methods that are accessing the pageStore are synchronized
// to allow deferred page creation (that is creation of pages
// while the document is already displayed.

- (unsigned int) countPages
{
   unsigned int count;
   
   [pageStoreLock lock];
   count = [pageStore count];
   [pageStoreLock unlock];

   return count;
}


/*
 * Get the page at a specified position. Note that the
 * first page has position 1 and the last page [self
 * countPages].
 */
- (ViewablePage*) pageAt:(unsigned int)index
{
   ViewablePage* thePage;

   NSAssert(((index > 0) && (index <= [pageStore count])),
            @"Page Index out of range");
   
   [pageStoreLock lock];
   thePage = [pageStore objectAtIndex: (index - 1)];
   [pageStoreLock unlock];

   return thePage;
}


- (void) appendPage:(ViewablePage*)aPage
{
   NSAssert(![pageStore containsObject:aPage],
            @"Page already containted in this Document");

   [pageStoreLock lock];

   [pageStore addObject: aPage];
   [aPage setPageNumber: [pageStore count]];

   NSLog(@"Page number %d added", [aPage pageNumber]);

   [pageStoreLock unlock];
}


/*
 * Must return YES if the document provides the ability
 * to adjust it's paper size. If so, the method setPapersize
 * is used by ViewIt to set the documents papersize when
 * desired. Also, the setPageOrientaton method has to be
 * implemented to do something meaningfull.
 *
 * This default implementation returns NO. You want to
 * overwrite this in subclasses if your document allows
 * papersize adjustion.
 */
- (BOOL) canSetPapersize
{
   return NO;
}


/*
 * Used to set the papersize if canSetPapersize returns
 * YES.
 *
 * This default implementation does nothing since the
 * default implementation of canSetPapersize returns
 * NO.
 */
- (void) setPapersize: (NSSize)papersize
{
   return;
}


/*
 * Used to set the page orientation if canSetPapersize
 * returns YES.
 *
 * This default implementation does nothing since
 * the default implementation of canSetPapersize returns
 * NO.
 */
- (void) setPageOrientation: (PageOrientation)orientation
{
   return;
}


/*
 * Invoked by the document's controller when the document
 * is displayed (visible to the user). The default behavior
 * is to do nothing.
 */
- (void) documentIsDisplayed
{
   return;
}


// NSDocument methods

- (void) close
{
   NSLog(@"Close Document");
   [super close];
}


/*
 * Create an instance of ImageWindowController that will
 * control a window that displays this document.
 */ 
- (void) makeWindowControllers
{
   [[AppController sharedController] createAndDisplayWindowForDocument: self];

   NSLog(@"new window controller created");   
}

@end
