
// created 06.2003 by Stefan Kleine Stegemann
// 
// licensed under GPL


#include "ViewablePDFDocument.h"
#include "ViewablePDFPage.h"
#include "PDFDocument.h"
#include "PDFRenderer.h"
#include "PDFRenderer2.h"
#include "PageCache.h"
#include "ViewIt.h"

#include <Foundation/NSString.h>
#include <Foundation/NSException.h>
#include <Foundation/NSUserDefaults.h>
#include <Foundation/NSValue.h>

/*
 * Non-Public methods.
 */
@interface ViewablePDFDocument(Private)
- (void) makeWindowControllers;
@end


/*
 * Makes a PDF document viewable through ViewIt.
 */
@implementation ViewablePDFDocument

- (id) init
{
   if (self = [super init])
   {
      pdfDoc   = nil;
      renderer = nil;
      cache    = nil;
   }

   return self;
}


- (void) dealloc
{
   NSLog(@"dealloc ViewablePDFDocument, retain count is %d", [self retainCount]);

   [cache release];
   [renderer release];
   [pdfDoc release];

   [super dealloc];
}


- (BOOL) readFromFile: (NSString*)fileName ofType: (NSString*)docType
{
   int               i;
   ViewablePDFPage*  aPage;
   BOOL              succeeded = YES;
   NSUserDefaults*   defaults = [NSUserDefaults standardUserDefaults];

   pdfDoc = [[PDFDocument documentFromFile: fileName] retain];
   if ([pdfDoc isOk])
   {
      NSString* gsExec = [defaults objectForKey: PREFS_GS_EXECUTABLE];
      int       cacheSize = [[defaults objectForKey: PREFS_CACHE_SIZE] intValue];
      int       prefetchCount = [[defaults objectForKey: PREFS_PREFETCH_COUNT] intValue];

      renderer = [[PDFRenderer alloc] initWithDocument: pdfDoc
                                      xRes: 144 
                                      yRes: 144
                                      gsExecutable: gsExec];
      [renderer setScale: 1.0];

      cache = [[PageCache alloc] initWithRenderer: renderer 
                                 cacheSize: cacheSize
                                 prefetchCount: prefetchCount
                                 notificationObject: nil];

      for (i = 1; i <= [pdfDoc numPages]; i++)
      {
         aPage = [[ViewablePDFPage alloc] initWithDocument: self pageNumber: i];
         [aPage autorelease];
         [self appendPage: aPage];
      }
   }
   else
   {
      succeeded = NO;
   }

   return succeeded;
}


/*
 * Returns the cache for page images. This method is
 * intended to be used by ViewablePDFPage.
 */
- (PageCache*) cache
{
   return cache;
}


/*
 * Returns the renderer to be used to create
 * images for the document pages.
 */
- (id) renderer
{
   return renderer;
}


/*
 * Returns the PDFDocument instance that is associated to
 * this ViewableDocument.
 */
- (id) pdfDoc
{
   return pdfDoc;
}


/*
 * Set the scale factor for the pages of the
 * document.
 */
- (void) setScaleFactor: (double)factor
{
   NSAssert(renderer != nil, @"no renderer");
   NSAssert(cache != nil, @"no cache");

   if ([renderer setScale: factor])
   {
      [cache invalidate];
   }
}


/*
 * Returns YES because the papersize can be
 * set for this type of document.
 */
- (BOOL) canSetPapersize
{
   return YES;
}


/*
 * Sets the papersize of this document to
 * the specified size.
 */
- (void) setPapersize: (NSSize)papersize
{
   NSAssert(renderer != nil, @"no renderer");
   NSAssert(cache != nil, @"no cache");

   if ([renderer setPaperSize: papersize])
   {
      [cache invalidate];
   }
}


/*
 * Sets the page orientation for this document.
 */
- (void) setPageOrientation: (PageOrientation)orientation
{
   PDFRPageOrientation po;

   NSAssert(renderer != nil, @"no renderer");
   NSAssert(cache != nil, @"no cache");

   po = (orientation == PageOrientation_Portrait ?
         PDFROrientation_Portrait : PDFROrientation_Landscape);

   if ([(PDFRenderer*)renderer setPageOrientation: po])
   {
      [cache invalidate];
   }
}


/*
 * Start caching when the document is displayed.
 */
- (void) documentIsDisplayed
{
   NSAssert(cache != nil, @"no cache");
   [cache startCaching];
}

@end
