
// created 06.2003 by Stefan Kleine Stegemann
// 
// licensed under GPL

#include "CacheLoader.h"
#include "PDFRenderer.h"
#include "PDFDocument.h"
#include "PageCache.h"

#include <Foundation/NSString.h>
#include <Foundation/NSException.h>
#include <Foundation/NSLock.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSValue.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSNotification.h>
#include <AppKit/NSImage.h>


#define THREAD_SYNC_UNDEF       0
#define THREAD_SYNC_WAIT_START  1
#define THREAD_SYNC_STARTED     2
#define THREAD_SYNC_IDLE        3
#define THREAD_SYNC_CONTINUE    4
#define THREAD_SYNC_TERMINATED  5


/*
 * Non-Public methods.
 */
@interface CacheLoader(Private)
- (void) loadPagesToCache;
@end


/*
 * An instance of this class runs is a separate thread
 * and loads the pages to be cached into the cache. Note
 * that this class is intended to be used internal only.
 */
@implementation CacheLoader

/*
 * Designated initalizer.
 */
- (id) initWithCache: (PageCache*)theCache 
{
   if (self = [super init])
   {
      cache            = theCache;
      isRunning        = NO;
      interruptCaching = NO;
      continueUpdate   = YES;
      cacheBasePage    = -1;
      syncLock         = [[NSConditionLock alloc] initWithCondition: THREAD_SYNC_UNDEF];
      startLock        = [[NSLock alloc] init];
      [startLock lock];
   }

   return self;
}


- (void) dealloc
{
   NSLog(@"dealloc CacheLoader, retain count is %d", [self retainCount]);

   [syncLock release];
   [startLock release];

   [super dealloc];
}


/*
 * Continuesly update the cache if needed. This
 * method has to be executed in a separate thread
 * and will fetch the pages before and after the
 * current page.
 */
- (void) run: (id)threadParam
{
   NSAutoreleasePool* threadPool;

   threadPool = [[NSAutoreleasePool alloc] init];

   NSLog(@"CacheLoader: waiting for start");
   [syncLock lock];

   [startLock lock];
   [startLock unlock];
   isRunning = YES;

   NSLog(@"CacheLoader: STARTED");
   [syncLock unlockWithCondition: THREAD_SYNC_IDLE];

   while (continueUpdate)
   {
      NSLog(@"CacheLoader: waiting for work");
      [syncLock lockWhenCondition: THREAD_SYNC_CONTINUE];
      
      [self loadPagesToCache];
      
      if (continueUpdate)
      {
         NSLog(@"CacheLoader: going to idle");
         [syncLock unlockWithCondition: THREAD_SYNC_IDLE];
      }

      [threadPool release];
      threadPool = [[NSAutoreleasePool alloc] init];
   }

   NSLog(@"CacheLoader: terminating");
   // FIXME: do something to ensure that the syncLock is locked
   [syncLock unlockWithCondition: THREAD_SYNC_TERMINATED];

   NSLog(@"CacheLoader: STOPPED");

   [threadPool release];
}


/*
 * Load the pages to cache.
 */
- (void) loadPagesToCache
{
   int                   nextPage;
   int                   direction;
   int                   distance;
   int                   count;
   NSImage*              pageImage;
   NSMutableDictionary*  notificationDict;

   NSAssert(cache != nil, @"no cache");

   notificationDict = [NSMutableDictionary dictionaryWithCapacity: 2];
   [notificationDict setObject: [NSNumber numberWithInt: [cache prefetchCount]]
                     forKey: CACHE_PROGRESS_INFO_PREFETCH_COUNT];

   distance  = 1;
   direction = 1;
   count     = 0;

   nextPage = cacheBasePage + (distance * direction);

   if ([cache notificationObject] && (!interruptCaching))
   {
      [notificationDict setObject: [NSNumber numberWithInt: count]
                        forKey: CACHE_PROGRESS_INFO_PAGES_DONE];
               
      [[NSNotificationCenter defaultCenter]
         postNotificationName: CACHE_PROGRESS_NOTIFICATION
         object: [cache notificationObject]
               userInfo: notificationDict];
   }


   while (continueUpdate &&
          (distance <= [cache prefetchCount]) &&
          (!interruptCaching))
   {
      [cache acquireCacheAccess];

      if ((nextPage >= 1) &&
          (nextPage <= [[[cache renderer] document] numPages]))
      {
         count++;

         if (![cache containsImageForPage: nextPage])
         {
            NSLog(@"PageCache-UpdateThread: loading page %d into the cache",
                  nextPage);
            [cache setNextHandledPageNumber: nextPage withBasePage: cacheBasePage];
            [[cache renderer] renderPage: nextPage];
         }
         else
         {
            NSLog(@"PageCache-UpdateThread: page %d is already cached", nextPage);
         }
         
         if ([cache notificationObject] && (!interruptCaching))
         {
            [notificationDict setObject: [NSNumber numberWithInt: count]
                              forKey: CACHE_PROGRESS_INFO_PAGES_DONE];
               
            [[NSNotificationCenter defaultCenter]
               postNotificationName: CACHE_PROGRESS_NOTIFICATION
               object: [cache notificationObject]
               userInfo: notificationDict];
         }
      }

      direction = (direction > 0 ? -1 : 1); // toggle direction
      if (direction == 1)
      {
         // increase distance if pages (basePage +/- distance)
         // have been loaded
         distance++;
      }

      nextPage = cacheBasePage + (distance * direction);

      [cache releaseCacheAccess];
   }
}


- (BOOL) isRunning
{
   return isRunning;
}


- (void) interruptCachingAsSoonAsPossible
{
   interruptCaching = YES;
}


- (int) cacheBasePage
{
   return cacheBasePage;
}


- (void) start
{
   NSAssert(![self isRunning], @"cache loader is already running");
   [startLock unlock];
}


- (void) continueWithBasePage: (int)basePage;
{
   if ([self isRunning])
   {
      [self interruptCachingAsSoonAsPossible];
      
      [syncLock lockWhenCondition: THREAD_SYNC_IDLE];
      interruptCaching = NO;
      cacheBasePage = basePage;
      [syncLock unlockWithCondition: THREAD_SYNC_CONTINUE];
   }
   else
   {
      interruptCaching = NO;
      cacheBasePage = basePage;
   }
}


- (void) terminate
{
   if ([self isRunning])
   {
      [self interruptCachingAsSoonAsPossible];
      [syncLock lockWhenCondition: THREAD_SYNC_IDLE];
      continueUpdate = NO;
      [syncLock unlockWithCondition: THREAD_SYNC_CONTINUE];
      NSLog(@"waiting for CacheLoader to terminate");
      [syncLock lockWhenCondition: THREAD_SYNC_TERMINATED];
      [syncLock unlockWithCondition: THREAD_SYNC_UNDEF];
   }
   else
   {
      continueUpdate = NO;
      [startLock unlock];
   }
}

@end
