
// created 06.2003 by Stefan Kleine Stegemann
// 
// licensed under GPL

#include "Ghostscript.h"

#include <Foundation/NSException.h>
#include <Foundation/NSString.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSTask.h>
#include <Foundation/NSFileHandle.h>
#include <Foundation/NSFileManager.h>
#include <Foundation/NSData.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSThread.h>
#include <Foundation/NSLock.h>
#include <Foundation/NSDate.h>
#include <Foundation/NSValue.h>
#ifdef MACOSX
#include <Foundation/NSPathUtilities.h>
#endif
#include <AppKit/NSImage.h>

#include <unistd.h>
#include <string.h>


/* for testing purpose only */
@interface GhostscriptImage : NSImage
{
}

- (void) dealloc;
@end

@implementation GhostscriptImage
- (void) dealloc
{
   NSLog(@"dealloc GhostscriptImage, retain count is %d", [self retainCount]);
   [super dealloc];
}
@end



/* counter to make temporary filenames unique */
static long gs_tmpfile_counter;

/*
 * Non-Public methods.
 */
@interface Ghostscript(Private)
- (NSString*) makeTempFileName: (NSString*)extension;
- (void) startGsIfNeeded;
- (void) stopGs;
- (void) readGsOut: (NSFileHandle*)output;
@end

/*
 * Provides an interface to a ghoscript interpreter. You
 * can send data to the interpreter and will get the rendered
 * result as images.
 *
 * Note that this implementation is currently not thread safe.
 */
@implementation Ghostscript

/*
 * Designated initialzer.
 */
- (id) initWithGsExecutable: (NSString*)pathToGsExecutable;
{
   if (self = [super init])
   {
      tempFileName     = [[self makeTempFileName: @"tiff"] retain];
      gsTask           = nil;
      toGsPipe         = nil;
      toGs             = nil;
      fromGsPipe       = nil;
      fromGs           = nil;
      needStartGs      = YES;
      xRes             = 0;
      yRes             = 0;
      gsAccessLock     = [[NSLock alloc] init];
      gsExecutable     = [pathToGsExecutable retain];

      [self setXRes: DEFAULT_X_RESOLUTION yRes: DEFAULT_Y_RESOLUTION];
      [self setPageSize: NSMakeSize(0, 0)];
   }

   return self;
}


- (void) dealloc
{
   NSLog(@"dealloc Ghostscript, retain count is %d", [self retainCount]);

   [tempFileName release];
   [gsAccessLock release];
   // just to be sure the data is really released
   [gsExecutable release];

   [super dealloc];
}


/*
 * This method has to be called when this Ghoscript
 * instance is no longer needed before the instance
 * is released. It will take care that the ghostscript
 * intepreter is stopped (and thus the output reading
 * thread which also retained this instance).
 */
- (void) prepareForTermination
{
   [self stopGs];
}


/*
 * Class initialization.
 */
+ (void) initialize
{
   [super initialize];
   gs_tmpfile_counter = 1;
}


/*
 * Set the resolution of the produced output. If the
 * specified resolution differs from the current one,
 * the gs interpreter will be restarted. This will discard
 * all pages that are currently in memory.
 */
- (void) setXRes: (int)newXRes yRes: (int)newYRes
{
   if (([self xRes] != newXRes) || ([self yRes] != newYRes))
   {
      NSLog(@"resolution changed, gs interpreter needs to be restarted");
      [self stopGs];
   }

   xRes = newXRes;
   yRes = newYRes;
}


/*
 * Set the page-size for the produced output. By default
 * the page size is derived from the gs input.
 */
- (void) setPageSize: (NSSize)size
{
   if ((pageSize.width != size.width) || (pageSize.height != size.height))
   {
      NSLog(@"pagesize changed, gs interpreter needs to be restared");
      [self stopGs];
   }

   pageSize = size;
}


/*
 * Get the current page size. If a property (width or height)
 * is 0, it will be ignored.
 */
- (NSSize) pageSize
{
   return pageSize;
}


/*
 * Get the current resolution of the x axis.
 */
- (int) xRes
{
   return xRes;
}


/*
 * Get the current resolution on the y axis.
 */
- (int) yRes
{
   return yRes;
}


/*
 * Whenever the Ghostscript instance is used
 * to process data, this method has to be
 * invoked before to ensure that concurrent
 * access is synchronized. This method will
 * block until access is granted.
 */
- (void) acquireAccess
{
   NSLog(@"waiting for access to ghostscript interpreter ...");
   [gsAccessLock lock];
   NSLog(@"access to ghostscript interpreter has been acquired");
}


/*
 * Free access to this Ghostscript instance a
 * fter the data has been processed and the instance
 * no longer needed to process data. Only call
 * this method when acquireAccess has been called
 * before.
 */
- (void) releaseAccess
{
   NSLog(@"releasing access to ghostscript interpreter");
   [gsAccessLock unlock];
}


/*
 * Pass data to the ghostscript interpreter.
 */
- (void) sendData: (NSData*)data
{
   [self startGsIfNeeded];

   NSAssert(toGs != nil, @"no pipe to gs interpreter");

   //NSLog(@"sending %d bytes of data to gs interpreter", [data length]);
   [toGs writeData: data];
}


/*
 * Send a command directly to the ghostscript interpreter.
 * A newline is automatically appended to the command.
 */
- (void) sendCommand: (NSString*)aCommand
{
   NSString*   cmdWithNl;
   NSData*     data;
   const char* bytes;

   [self startGsIfNeeded];
   
   NSAssert(toGs != nil, @"to pipe to gs interpreter");
   NSLog(@"sending command <%@> to the gs interpreter", aCommand);

   cmdWithNl = [NSString stringWithFormat: @"%@\n", aCommand];
   bytes = [cmdWithNl UTF8String];
   data = [[NSData alloc] initWithBytes: bytes length: strlen(bytes)];
   [self sendData: data];
   [data release];
}


/*
 * Start the ghostscript interpreter if needed.
 */
- (void) startGsIfNeeded
{
   if (needStartGs)
   {
      NSMutableArray* args;

      NSLog(@"starting ghostscript interpreter ...");

      toGsPipe = [NSPipe pipe]; // retained/released by task below
      toGs     = [toGsPipe fileHandleForWriting];

      fromGsPipe = [NSPipe pipe]; // retained/released by task below
      fromGs     = [fromGsPipe fileHandleForReading];

      gsTask = [[NSTask alloc] init];
      NSLog(@"gs executable path is %@", gsExecutable);
      [gsTask setLaunchPath: gsExecutable];

      args = [NSMutableArray arrayWithCapacity: 1];
      [args addObject: @"-dNOPAUSE"];
      [args addObject: [NSString stringWithFormat: @"-sOutputFile=%@",
                                 tempFileName]];
      [args addObject: [NSString stringWithFormat: @"-r%dx%d",
                                 [self xRes], [self yRes]]];
      [args addObject: @"-sDEVICE=tiff24nc"];
      [args addObject: @"-dTextAlphaBits=4"];
      [args addObject: @"-dQUIET"];
      [args addObject: @"-dSAFER"];

      /*
      if (pageSize.width != 0)
      {
         [args addObject: [NSString stringWithFormat: @"-dDEVICEWIDTH=%d",
                                    pageSize.width]];
      }
      if (pageSize.height != 0)
      {
         [args addObject: [NSString stringWithFormat: @"-dDEVICEHEIGHT=%d",
                                    pageSize.height]];
      }
      */

      [args addObject: @"-"];

      [gsTask setStandardInput: toGsPipe];
      [gsTask setStandardOutput: fromGsPipe];

      [gsTask setArguments: args];
      [gsTask launch];

      nextOutputPageIndex = 1;

      // TODO: read output in background

      NSLog(@"gs interpreter has been started, output goes to %@", tempFileName);
      
      needStartGs = NO;
   }
}


/*
 * Stop the ghostscript interpreter if it runs.
 * The corresponding task will be deallocated.
 */
- (void) stopGs
{
   if (gsTask != nil)
   {
      if ([gsTask isRunning])
      {
         [self acquireAccess];

         [self sendCommand: @"quit"];
         NSLog(@"ghostscript interpreter has been asked to stop");

         //NSLog(@"waiting for gs task to exit ...");
         //[gsTask waitUntilExit];
         //NSLog(@"gs task exitet ...");
         //[gsTask terminate];

         [self releaseAccess];
      }

      [gsTask release];
      
      gsTask     = nil;
      toGsPipe   = nil;
      toGs       = nil;
      fromGsPipe = nil;
      fromGs     = nil;

      needStartGs = YES;

      NSLog(@"ghostscript interpreter needs to be restarted");
   }
}


/*
 * Read the output from the ghostscript interpreter.
 */
- (void) readGsOut: (NSFileHandle*)output;
{
   NSData*            data;
   NSString*          msg;
   NSAutoreleasePool* threadPool;

   threadPool = [[NSAutoreleasePool alloc] init];

   NSAssert(output != nil, @"cannot read from nil handle");
   
   NSLog(@"gs output reader started");

   // read data from gs output
   [output retain]; 

   data = [output availableData];
   while ([data length] > 0)
   {
      msg = [[NSString alloc] initWithData: data
                              encoding: NSASCIIStringEncoding];
      NSLog(@"GS-OUTPUT: >%@<", msg);      
      [msg release];

      data = [output availableData];
   }

   // clean up
   [threadPool release];
   [output release];
   
   NSLog(@"gs output reader stopped");
}


/*
 * Wait until the next image has been produced by ghostscript
 * and do something with this image. Therefore, a handleGSOutput
 * message is send to the specified handler object. The object
 * returned by this handler will be returned by this method.
 *
 * To determine if an image is 'ready' this method uses to
 * 'two-way'-strategy. First, it waits until the file for
 * the image becomes available. Then it waits until it's
 * file size does not change any more and creates the image
 * from the file. If the produced image is not valid it is
 * assumed that the file is not complete and size-checking
 * is restarted. If the image could not be created after
 * 10 seconds, an exception is raised.
 */
- (id) nextImage: (id)handler
{
   int                 totalWaitTime;
   BOOL                imageReady;
   NSString*           nextExpectedFile;
   id                  result;
   unsigned long long  currSize;
   unsigned long long  newSize;
   NSFileManager*      fileMan = [NSFileManager defaultManager];

   nextExpectedFile = [NSString stringWithFormat: tempFileName,
                                nextOutputPageIndex];

   imageReady    = NO;
   totalWaitTime = 0;   
   currSize      = 0;

   while (!imageReady)
   {
      //NSLog(@"waiting for output file, curren size is %ld ...", currSize);
      [NSThread sleepUntilDate: [NSDate dateWithTimeIntervalSinceNow: 0.050]];
                
      if ([fileMan fileExistsAtPath: nextExpectedFile])
      {
         newSize = [[fileMan fileAttributesAtPath: nextExpectedFile
                             traverseLink: NO] fileSize];

         //NSLog(@"   new size is %ld", newSize);

         if ((newSize == currSize) && (currSize > 0))
         {
            NSImage* image;

            image = [[GhostscriptImage alloc] initWithContentsOfFile: nextExpectedFile];
            if ((image == nil) || (![image isValid]))
            {
               NSLog(@"image corrupted, file seems not to be complete, retry");
            }
            else
            {
               imageReady = YES;
            }

            [image release];
         }
         else
         {
            currSize = newSize;
         }
      }
   }

   if (!imageReady)
   {
      [NSException raise: @"NoSuchImageFromGs" 
                   format: @"image %d not available", nextOutputPageIndex];
   }

   result = [handler handleGSOutput: nextExpectedFile];

   nextOutputPageIndex++;

   // delete the temporary file if not already done by the handler
   if ([fileMan fileExistsAtPath: nextExpectedFile])
   {
      if (![fileMan removeFileAtPath: nextExpectedFile handler: nil])
      {
         NSLog(@"cannot remove temporary file %@", nextExpectedFile);
      }
   }

   return result;
}


/*
 * Process the next count images. This is convience method
 * to avoid subsequent calls to nextImage while counting
 * the number of retrieved images in the application.
 */
- (NSArray*) nextImages: (int)count withHandler: (id)handler
{
   NSMutableArray* result;
   int             i;

   result = [[NSMutableArray alloc] initWithCapacity: count];
   for (i = 0; i < count; i++)
   {
      [result addObject: [self nextImage: handler]];
   }

   return [result autorelease];
}


/*
 * Create a (hopefully) unique filename. This is the file where
 * the ghoscript output goes to. If the specified extension is
 * non-nil it will be appended to the created filename.
 */
- (NSString*) makeTempFileName: (NSString*)extension
{
   NSString* tmpName;
   NSString* tmpPath;
   int       fileCounter;

   fileCounter = gs_tmpfile_counter++;
   tmpName = 
      [[[NSString alloc] initWithFormat: @"gs%ld_%d_%%d",
                        getpid(),
                        fileCounter] autorelease];

   tmpPath = 
      [(NSString*)NSTemporaryDirectory() stringByAppendingPathComponent: tmpName];

   if (extension != nil)
   {
      tmpPath = [tmpPath stringByAppendingPathExtension: extension];
   }

   return tmpPath;
}

@end
