
// created 06.2003 by Stefan Kleine Stegemann
// 
// licensed under GPL


#include "PDFDocument.h"
#include "XPDFBridge.h"

#include <Foundation/NSException.h>
#include <Foundation/NSString.h>

/*
 * Non-Public methods.
 */
@interface PDFDocument(Private)
- (id) initWithBridgeDocument: (XPDFObject)aPdfDoc;
@end

/*
 * Provides a wrapper for an XPDF PDFDoc object. Such
 * an object provides access to the meta-data of a
 * PDF document as well as to the document content.
 */
@implementation PDFDocument


- (id) init
{
   return [self initWithBridgeDocument: NULL];
}


/*
 * Designated initializer.
 */
- (id) initWithBridgeDocument: (XPDFObject)aPdfDoc
{
   if (self = [super init])
   {
      pdfDoc = aPdfDoc;
   }

   return self;
}   


- (void) dealloc
{
   NSLog(@"dealloc PDFDocument, retain count is %d", [self retainCount]);

   if (pdfDoc != NULL)
   {
      PDFDoc_delete(pdfDoc);
      pdfDoc = NULL;
   }
   else
   {
      NSLog(@"   no internal document to delete");
   }

   [super dealloc];
}


+ (PDFDocument*)documentFromFile: (NSString*)fileName
{
   return [PDFDocument documentFromFile: fileName
                       ownerPassword: nil
                       userPassword: nil];
}

+ (PDFDocument*)documentFromFile: (NSString*)fileName 
                   ownerPassword: (NSString*)ownerPassword
                    userPassword: (NSString*)userPassword
{
   XPDFObject   newPdfDoc;
   const char*  cFileName;
   const char*  cOwnerPassword;
   const char*  cUserPassword;

   NSAssert(fileName != nil, @"no filename");

   cFileName = [fileName cString];
   
   cOwnerPassword = (ownerPassword != nil ? [ownerPassword cString] : NULL);
   cUserPassword  = (userPassword != nil ? [userPassword cString] : NULL);

   newPdfDoc = PDFDoc_create(cFileName,
                             cOwnerPassword,
                             cUserPassword);

   return [[[PDFDocument alloc] initWithBridgeDocument: newPdfDoc] autorelease];
}


- (BOOL) isOk
{
   NSAssert(pdfDoc != NULL, @"no document");

   return PDFDoc_isOk(pdfDoc);
}

- (int) errorCode
{
   NSAssert(pdfDoc != NULL, @"no document");
   
   return PDFDoc_getErrorCode(pdfDoc);
}


- (double) pageWidth: (int)pageNum
{
   NSAssert(pdfDoc != NULL, @"no document");
   
   return PDFDoc_getPageWidth(pdfDoc, pageNum);
}


- (double) pageHeight: (int)pageNum
{
   NSAssert(pdfDoc != NULL, @"no document");
   
   return PDFDoc_getPageHeight(pdfDoc, pageNum);
}


- (int) pageRotate: (int)pageNum
{
   NSAssert(pdfDoc != NULL, @"no document");
   
   return PDFDoc_getPageRotate(pdfDoc, pageNum);
}


- (int) numPages
{
   NSAssert(pdfDoc != NULL, @"no document");
   
   return PDFDoc_getNumPages(pdfDoc);
}


- (NSString*)metaData
{
   const char* data;

   NSAssert(pdfDoc != NULL, @"no document");

   return (data != NULL ? 
           [[[NSString alloc] initWithCString: data] autorelease]
           : 
           nil);
}


/*
 * Returns the internal wrapped object.
 */
- (void*) xpdfobject
{
   return pdfDoc;
}


@end
