
// created 06.2003 by Stefan Kleine Stegemann
// 
// licensed under GPL

#include "PDFRenderer.h"
#include "PDFDocument.h"
#include "Ghostscript.h"
#include "XPDFBridge.h"

#include <Foundation/NSObject.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSString.h>
#include <Foundation/NSException.h>
#include <Foundation/NSFileManager.h>
#include <Foundation/NSData.h>
#include <AppKit/NSImage.h>

#include <unistd.h>


/*
 * Callback method to pass data to a Ghoscript Interpreter.
 * The argument stream is expected to point to an instance
 * of Ghostscript.
 */
static void outputToGs(void *stream, char* data, int len)
{
   Ghostscript* gs;
   NSData*      gsData;

   gs = (Ghostscript*)stream;

   gsData = [[NSData alloc] initWithBytes: data length: len];
   [gs sendData: gsData];
   [gsData release];
}


/*
 * Non-Public methods.
 */
@interface PDFRenderer(Private)
- (id) handleGSOutput: (NSString*)gsOutputFile;
@end



/*
 * Render pages of a PDFDocument. Each page is rendered by
 * passing it to an instance of Ghostscript. The output
 * produced by a PDFRenderer is varying depending on the
 * handler that is used to handle the Ghostscript output.
 * If no handler is specified, the PDFRenderer will handle
 * the output itself. This results in an image for each
 * rendered page.
 */
@implementation PDFRenderer

/*
 * Designated initializer. The results produced by this
 * renderer depend on the specified gsOutputHandler.
 */
- (id) initWithDocument: (PDFDocument*)pdfDoc 
                   xRes: (int)xResolution
                   yRes: (int)yResolution
           gsExecutable: (NSString*)gsExecutable
        gsOutputHandler: (id)handler
{
   if (self = [super init])
   {
      myDocument = [pdfDoc retain];
      
      baseXRes = xResolution;
      baseYRes = yResolution;

      gs = [[Ghostscript alloc] initWithGsExecutable: gsExecutable];
      [self setGsOutputHandler: self];
      [self setScale: 1.0];
      [self setPaperSize: NSMakeSize(0, 0)];
      [self setPageOrientation: PDFROrientation_Portrait];
   }

   return self;
}


/*
 * Create a new PDFRenderer that produces an image for
 * each rendered page.
 */
- (id) initWithDocument: (PDFDocument*)pdfDoc 
                   xRes: (int)xResolution
                   yRes: (int)yResolution
           gsExecutable: (NSString*)gsExecutable
{
   return [self initWithDocument: pdfDoc
                xRes: xResolution
                yRes: yResolution
                gsExecutable: gsExecutable
                gsOutputHandler: self];
}


- (void) dealloc
{
   NSLog(@"dealloc PDFRenderer, retain count is %d", [self retainCount]);

   [myDocument release];
   [gs prepareForTermination];
   [gs release];

   [super dealloc];
}


/*
 * Render a couple of pages. The method throws an exception
 * if something went wrong. The first page is 1.
 */
- (NSArray*) renderFirstPage: (int)firstPage
                    lastPage: (int)lastPage   
{
   RenderResult    renderRc;
   NSArray*        result;
   OutputOptions   options;


   NSAssert(myDocument != nil, @"no PDFDocument");
   NSAssert(gs != nil, @"no interface to ghostscript");
   NSAssert([myDocument xpdfobject] != NULL, @"invalid PDFDocument object");
   NSAssert([myDocument isOk], @"cannot render a document that is not ok");
   NSAssert((firstPage > 0) && (firstPage <= lastPage), 
            @"first page of page range is invalid");
   NSAssert(lastPage <= [myDocument numPages],
            @"last page of page range exceeds number of available pages");


   options.firstPage   = firstPage;
   options.lastPage    = lastPage;

   if (paperSize.width > 0)
   {
      options.paperWidth  = paperSize.width;
   }
   else
   {
      options.paperWidth = [myDocument pageWidth: firstPage] + 0.5;
   }

   if (paperSize.height > 0)
   {
      options.paperHeight = paperSize.height;
   }
   else
   {
      options.paperHeight = [myDocument pageHeight: firstPage] + 0.5;
   }

   if ([self pageOrientation] == PDFROrientation_Landscape)
   {
      int width = options.paperWidth;
      options.paperWidth  = options.paperHeight;
      options.paperHeight = width;
   }

   options.resolution  = 72;

   [gs acquireAccess];
   renderRc = Renderer_render([myDocument xpdfobject],
                              outputToGs,
                              (void*)gs,
                              &options);
   [gs releaseAccess];

   if (renderRc != RenderResult_Ok)
   {
      [NSException raise: @"PDFRenderingFailed" 
                   format: @"rendering failed (rc=%d)", renderRc];
   }

   result = [gs nextImages: ((lastPage - firstPage) + 1) withHandler: gsOutputHandler];
   NSLog(@"%d images returned by ghostscript", [result count]);
   NSAssert([result count] == ((lastPage - firstPage) + 1),
            @"invalid number of images returned by Ghostscript");

   return result;
}


/*
 * Render one page.
 */
- (id) renderPage: (int)page
{
   return (NSImage*)[[self renderFirstPage:page 
                           lastPage:page] objectAtIndex: 0];
}


/*
 * Render the whole document. The method throws
 * an exception if something went wrong. The first
 * page is 1.
 */
- (NSArray*) render
{
   return [self renderFirstPage: 1 lastPage: [myDocument numPages]];
}


/*
 * Set the handler for handling Ghostscript output.
 */
- (void) setGsOutputHandler: (id)handler
{
   gsOutputHandler = handler;
}


/*
 * Set the Factor for the following renderings based
 * on the default resolution that has been given to
 * the initializer. Returns true if the new scale
 * factor is different to the current.
 */
- (BOOL) setScale: (float)newScale
{
   BOOL changed;

   NSAssert(gs != nil, @"no interface to ghostscript");

   if (scale != newScale)
   {
      scale = newScale;
      [gs setXRes: (int)(baseXRes * scale)  yRes: (int)(baseYRes * scale)];
      changed = YES;
   }
   else
   {
      changed = NO;
   }

   return changed;
}


/*
 * Get the current scale factor.
 */
- (float) scale
{
   return scale;
}


/*
 * Set the paper size for the rendered pages. If one
 * of the properties (width or height) is 0, the default
 * for this property will be used. Returns YES if
 * the new papersize differs from the current size.
 */
- (BOOL) setPaperSize: (NSSize)size
{
   BOOL changed = NO;

   if ((paperSize.width != size.width) || (paperSize.height != size.height))
   {
      paperSize = size;
      changed = YES;
   }

   return changed;
}


/*
 * Get the current paper size. 0 for width or height
 * means use default for this property.
 */
- (NSSize) paperSize
{
   return paperSize;
}


/*
 * Set the page orientation for rendered pages. If
 * the specified orientation is different from the
 * current one, YES is returned; otherwise this method
 * returns NO.
 */
- (BOOL) setPageOrientation: (PDFRPageOrientation)orientation
{
   BOOL changed = NO;

   if (pageOrientation != orientation)
   {
      pageOrientation = orientation;
      changed = YES;
   }

   return changed;
}


/*
 * Returns the orientation used for rendered pages.
 */
- (PDFRPageOrientation) pageOrientation
{
   return pageOrientation;
}


/*
 * Get the document that is processed by this
 * renderer.
 */
- (PDFDocument*) document
{
   return myDocument;
}


/*
 * Handle output produced by the Ghostscript interpreter.
 * This implementation will produce and returned an image 
 * from this output.
 */
- (id) handleGSOutput: (NSString*)gsOutputFile
{
   return [[[NSImage alloc] initWithContentsOfFile: gsOutputFile] autorelease];
}


@end
