
// created 06.2003 by Stefan Kleine Stegemann
// 
// licensed under GPL

#include "PDFRenderer2.h"
#include "PDFDocument.h"
#include "PDFImageRep.h"

#include <Foundation/NSObject.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSString.h>
#include <Foundation/NSException.h>
#include <AppKit/NSImage.h>

/*
 * Non-Public methods.
 */
@interface PDFRenderer2(Private)
@end


/*
 * Render pages of a PDFDocument. The renderer produces
 * an image for each page.
 */
@implementation PDFRenderer2

/*
 * Designated initializer.
 */
- (id) initWithDocument: (PDFDocument*)pdfDoc 
                   xRes: (int)xResolution
                   yRes: (int)yResolution
{
   if (self = [super init])
   {
      myDocument = [pdfDoc retain];
      
      baseXRes = xResolution;
      baseYRes = yResolution;

      [self setScale: 1.0];
      [self setPaperSize: NSMakeSize(0, 0)];
   }

   return self;
}


- (void) dealloc
{
   NSLog(@"dealloc PDFRenderer2, retain count is %d", [self retainCount]);

   [myDocument release];

   [super dealloc];
}


/*
 * Render a couple of pages. The returned array
 * is autoreleased and contains an image for each
 * document page. The method throws an exception
 * if something went wrong. The first page is 1.
 */
- (NSArray*) renderFirstPage: (int)firstPage
                    lastPage: (int)lastPage
{
   NSMutableArray*  result;
   int              pageCount;
   int              numPages;

   NSAssert(myDocument != nil, @"no PDFDocument");
   NSAssert([myDocument isOk], @"cannot render a document that is not ok");
   NSAssert((firstPage > 0) && (firstPage <= lastPage), 
            @"first page of page range is invalid");
   NSAssert(lastPage <= [myDocument numPages],
            @"last page of page range exceeds number of available pages");


   numPages = (lastPage - firstPage) + 1;
   result = [[[NSMutableArray alloc] initWithCapacity: numPages] autorelease];

   for (pageCount = firstPage; pageCount <= lastPage; pageCount++)
   {
      NSImage*     anImage;
      NSSize       imageSize;
      PDFImageRep* imageRep;

      imageSize = NSMakeSize([myDocument pageWidth: pageCount],
                             [myDocument pageHeight: pageCount]);

      imageRep = [[[PDFImageRep alloc] init] autorelease];
      [imageRep setSize: imageSize];

      anImage = [[NSImage alloc] initWithSize: imageSize];
      [anImage addRepresentation: imageRep];
      
      [result addObject: anImage];
      [anImage autorelease];
   }

   return result;
}


/*
 * Render one page. The returned image is autoreleased.
 * The method throws an exception if something went
 * wrong.
 */
- (NSImage*) renderPage: (int)page
{
   return (NSImage*)[[self renderFirstPage:page 
                           lastPage:page] objectAtIndex: 0];
}


/*
 * Render the whole document. The returned array
 * is autoreleased and contains an image for each
 * document page. The method throws an exception
 * if something went wrong. The first page is 1.
 */
- (NSArray*) render
{
   return [self renderFirstPage: 1 lastPage: [myDocument numPages]];
}


/*
 * Set the Factor for the following renderings based
 * on the default resolution that has been given to
 * the initializer. Returns true if the new scale
 * factor is different to the current.
 */
- (BOOL) setScale: (float)newScale
{
   BOOL changed;

   if (scale != newScale)
   {
      scale = newScale;
      changed = YES;
   }
   else
   {
      changed = NO;
   }

   return changed;
}


/*
 * Get the current scale factor.
 */
- (float) scale
{
   return scale;
}


/*
 * Set the paper size for the rendered pages. If one
 * of the properties (width or height) is 0, the default
 * for this property will be used.
 */
- (BOOL) setPaperSize: (NSSize)size
{
   paperSize = size;
   return YES;
}


/*
 * Get the current paper size. 0 for width or height
 * means use default for this property.
 */
- (NSSize) paperSize
{
   return paperSize;
}


/*
 * Get the document that is processed by this
 * renderer.
 */
- (PDFDocument*) document
{
   return myDocument;
}


@end
