
// created 06.2003 by Stefan Kleine Stegemann
// 
// licensed under GPL

#include "PageCache.h"
#include "CacheLoader.h"
#include "PDFDocument.h"
#include "PDFRenderer.h"

#include <Foundation/NSString.h>
#include <Foundation/NSException.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSEnumerator.h>
#include <Foundation/NSDictionary.h>
#include <Foundation/NSLock.h>
#include <Foundation/NSValue.h>
#include <Foundation/NSThread.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSFileManager.h>

#include <stdlib.h>


/* counter to make temporary filenames unique */
static long pc_tmpfile_counter;

/*
 * Non-Public methods.
 */
@interface PageCache(Private)
- (id) handleGSOutput: (NSString*)gsOutputFile;
- (NSString*) makeTempFileName;
- (void) removeImageFile: (NSString*)imageFile;
@end


/*
 * This cache is intended to increase the performance
 * when displaying or working with rendered documents.
 * When a page is accessed (imageForPage), the cache
 * will provident attempt to load a number of pages 
 * around the accessed page. The amount of pages is
 * configurable and has to be specified to the initializer.
 * Note that eg. specifying an amount of 2 will cause
 * the cache to load 4 pages at maximum (2 pages before
 * and 2 pages ahead).
 *
 * Note that caching is deferred. Do initiate caching,
 * you have to call the startCaching method. This has
 * been done because it is desirable to display the
 * first page as soon as possible, therefore seems
 * not a good idea to have caching working in the background
 * while the first page is loaded.
 */
@implementation PageCache

/*
 * Designated initialzer. cacheSize specifies the number
 * of pages that will be loaded providently. See class
 * description for more information. If notificationObject
 * is non-nil and the cache is configured to preload pages,
 * notifcations will be sent for this object about the
 * preloading-progress. Thus, observers that register
 * for this Object an the CACHE_PROGRESS_NOTIFICATION
 * will receive those notifications. The notification
 * object is not retained.
 */
- (id) initWithRenderer: (id)aRenderer 
              cacheSize: (int)maxPagesInCache
          prefetchCount: (int)numPages
     notificationObject: (id)anObject
{
   if (self = [super init])
   {
      NSAssert(maxPagesInCache > 0, @"cache size has to be greater than 0");

      renderer            = [aRenderer retain];
      cacheAccessLock     = [[NSLock alloc] init];
      cacheSize           = maxPagesInCache;
      prefetchCount       = numPages;
      notificationObject  = anObject;

      // we handle the gs output by ourselves
      [renderer setGsOutputHandler: self];
      [self setNextHandledPageNumber: -1 withBasePage: -1];

      cache = 
         [[NSMutableDictionary alloc] initWithCapacity: cacheSize];

      cacheLoader = [[CacheLoader alloc] initWithCache: self];

      [NSThread detachNewThreadSelector: @selector(run:)
                               toTarget: cacheLoader
                             withObject: nil];
   }

   return self;
}


- (void) dealloc
{
   NSLog(@"dealloc PageCache, retain count is %d", [self retainCount]);

   [cacheLoader terminate];

   [self invalidate]; // this will remove all temporary files

   [cacheLoader release];
   [cacheAccessLock release];
   [renderer release];
   [cache release];

   [super dealloc];
}


/*
 * Class initialization.
 */
+ (void) initialize
{
   [super initialize];
   pc_tmpfile_counter = 1;
}


/*
 * Start caching if not already started.
 */
- (void) startCaching
{
   NSAssert(cacheLoader != nil, @"no cache loader");

   if (![cacheLoader isRunning])
   {
      [cacheLoader start];
   }
}


- (NSImage*) imageForPage: (int)pageNum
{
   NSString*  imageFile;
   NSImage*   theImage;
   NSNumber*  key;

   NSLog(@"PageCache: waiting for cache access lock");
   [self acquireCacheAccess];
   NSLog(@"PageCache: got cache access lock");

   key = [NSNumber numberWithInt: pageNum];
   imageFile = [cache objectForKey: key];
   if (imageFile == nil)
   {
      [self setNextHandledPageNumber: pageNum withBasePage: pageNum];
      [renderer renderPage: pageNum];
      imageFile = [cache objectForKey: key];
   }
   else
   {
      NSLog(@"image can be read from cache (file %@)", imageFile);
   }

   NSAssert(imageFile != nil, @"image not found");

   theImage = [[[NSImage alloc] initWithContentsOfFile: imageFile] autorelease];

   [self releaseCacheAccess];

   if (pageNum != [cacheLoader cacheBasePage])
   {
      [cacheLoader continueWithBasePage: pageNum];
   }

   return theImage;
}


- (int) cacheSize
{
   return cacheSize;
}


- (int) prefetchCount
{
   return prefetchCount;
}


- (id) renderer
{
   return renderer;
}


/*
 * Returns the object to send progress notifications for
 * (see init-documentation). nil if no such object has been
 * set. In this case, no  notifications should be send.
 */
- (id) notificationObject
{
   return notificationObject;
}


- (void) acquireCacheAccess
{
   [cacheAccessLock lock];
}


- (void) releaseCacheAccess
{
   [cacheAccessLock unlock];
}


/*
 * Set the number of the page that will be send
 * to the gs output handler method next.
 */
- (void) setNextHandledPageNumber: (int)pageNum withBasePage: (int)basePageNum
{
   nextPageNumber      = pageNum;
   basePageForNextPage = basePageNum;
}


/*
 * Handle output produced by ghostscript.
 */
- (id) handleGSOutput: (NSString*)gsOutputFile
{
   NSFileManager* fileMan   = [NSFileManager defaultManager];
   NSString*      cacheFile = [self makeTempFileName];

   NSAssert(![fileMan fileExistsAtPath: cacheFile], @"cache file exists");
   NSAssert(nextPageNumber > 0, @"unexpected page");
   NSAssert(basePageForNextPage > 0, @"no base page for next page");

   if (![fileMan movePath: gsOutputFile 
                 toPath: cacheFile
                 handler: nil])
   {
      [NSException raise: @"PageCacheError"
                   format: @"cannot create cache file %@", cacheFile];
   }

   [self addToCache: cacheFile
         pageNum: nextPageNumber
         basePage: basePageForNextPage];

   [self setNextHandledPageNumber: -1 withBasePage: -1];

   return cacheFile;
}


/*
 * Add a new page image to the cache. When the number
 * images in the cache exceeds the cacheSize the images
 * for the pages with the highest distance to the 
 * specified basePage are removed.
 *
 * Note: must be executed with cache access acquired.
 */
- (void) addToCache: (NSString*)imageFile
            pageNum: (int)num
           basePage: (int)page
{
   NSEnumerator* e;
   id            aKey;
   id            keyToRemove;
   int           maxDistance;
   int           distance;
   id            object;


   while ([cache count] >= [self cacheSize])
   {
      maxDistance = -1;
      
      e = [cache keyEnumerator];
      while (aKey = [e nextObject])
      {
         distance = abs(page - [aKey intValue]);

         if (distance > maxDistance)
         {
            maxDistance = distance;
            keyToRemove = aKey;
         }
      }

      object = [cache objectForKey: keyToRemove];
      NSLog(@"PageCache: remove image for page %d", [keyToRemove intValue]);
      [self removeImageFile: [cache objectForKey: keyToRemove]];
      [cache removeObjectForKey: keyToRemove];
   }

   [cache setObject: imageFile forKey: [NSNumber numberWithInt: num]];
}


/*
 * Note: must be executed with cache access acquired.
 */
- (BOOL) containsImageForPage: (int)page
{
   NSNumber* key;

   key = [NSNumber numberWithInt: page];
   return ([cache objectForKey: key] != nil);
}


- (void) invalidate
{
   NSEnumerator* e;
   NSString*     anImageFile;

   [self acquireCacheAccess];
   
   e = [cache objectEnumerator];
   while (anImageFile = [e nextObject])
   {
      [self removeImageFile: anImageFile];
   }
   [cache removeAllObjects];

   [self releaseCacheAccess];
}


- (void) removeImageFile: (NSString*)imageFile
{
   NSFileManager* fileMan = [NSFileManager defaultManager];

   if ([fileMan fileExistsAtPath: imageFile])
   {
      if (![fileMan removeFileAtPath: imageFile handler: nil])
      {
         NSLog(@"cannot remove cache file %@", imageFile);
      }
   }
   else
   {
      NSLog(@"cache file %@ does not exist", imageFile);
   }
}


/*
 * Create a (hopefully) unique filename.
 */
- (NSString*) makeTempFileName
{
   NSString* tmpName;
   NSString* tmpPath;
   int       fileCounter;

   fileCounter = pc_tmpfile_counter++;
   tmpName = 
      [[[NSString alloc] initWithFormat: @"pagecache%ld_%d.tiff",
                        getpid(),
                        fileCounter] autorelease];

   tmpPath = 
      [(NSString*)NSTemporaryDirectory() stringByAppendingPathComponent: tmpName];

   return tmpPath;
}

@end
